#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for pyvo.registry.vao
"""


import os, sys, shutil, re, imp
import unittest, pdb
from urllib.error import URLError, HTTPError

from ...dal import query as dalq
from .. import vao as reg
from astropy.io.votable.tree import VOTableFile
from ...dal.query import _votableparse as votableparse
from astropy.utils.data import get_pkg_data_filename

regresultfile = "data/reg-short.xml"


class RegServiceTest(unittest.TestCase):

    def testCtor(self):
        self.res = {"title": "VAO Registry", "shortName": "vaoreg"}
        self.srv = reg.RegistryService(resmeta=self.res)

    def testProps(self):
        self.testCtor()
        self.assertEqual(self.srv.baseurl, 
                          reg.RegistryService.STSCI_REGISTRY_BASEURL)
        try:
            self.srv.baseurl = "goober"
            self.fail("baseurl not read-only")
        except AttributeError:
            pass

        self.assertEqual(self.srv.info["title"], "VAO Registry")
        self.assertEqual(self.srv.info["shortName"], "vaoreg")
        self.srv.info["title"] = "Sir"
        self.assertEqual(self.res["title"], "VAO Registry")

    def testCreateQuery(self):
        self.testCtor()
        q = self.srv.create_query()
        self.assertTrue(isinstance(q, reg.RegistryQuery))
        self.assertEqual(q.baseurl, reg.RegistryService.STSCI_REGISTRY_BASEURL)
        self.assertEqual(len(list(q._param.keys())), 0)

    def testCreateQueryWithArgs(self):
        self.testCtor()
        q = self.srv.create_query(keywords=["galaxy", "AGN"], servicetype="sia",
                                  waveband="radio", 
                                  sqlpred="publisher like '%nrao%'")
        self.assertTrue(isinstance(q, reg.RegistryQuery))
        self.assertEqual(q.baseurl, reg.RegistryService.STSCI_REGISTRY_BASEURL)
        self.assertEqual(len(list(q._param.keys())), 0)

        self.assertEqual(q.waveband, "Radio")
        self.assertEqual(q.servicetype, "sia")
        self.assertEqual(len(q.keywords), 2)
        self.assertEqual(q.keywords[0], "galaxy")
        self.assertEqual(q.keywords[1], "AGN")

        self.assertEqual(len(q.predicates), 1)
        self.assertEqual(q.predicates[0], "publisher like '%nrao%'")

    def testCreateQueryWithOrkw(self):
        self.testCtor()

        q = self.srv.create_query(keywords=["galaxy", "AGN"])
        self.assertTrue(isinstance(q, reg.RegistryQuery))
        self.assertEqual(q.baseurl, reg.RegistryService.STSCI_REGISTRY_BASEURL)
        self.assertEqual(len(list(q._param.keys())), 0)

        self.assertEqual(len(q.keywords), 2)
        self.assertEqual(q.keywords[0], "galaxy")
        self.assertEqual(q.keywords[1], "AGN")
        self.assertFalse(q.will_or_keywords())

        qurl = q.getqueryurl()
        self.assertTrue("+AND+%28title+" in qurl)
        
        q = self.srv.create_query(keywords=["galaxy", "AGN"], orkw=True)
        self.assertTrue(isinstance(q, reg.RegistryQuery))
        self.assertEqual(q.baseurl, reg.RegistryService.STSCI_REGISTRY_BASEURL)
        self.assertEqual(len(list(q._param.keys())), 0)

        self.assertEqual(len(q.keywords), 2)
        self.assertEqual(q.keywords[0], "galaxy")
        self.assertEqual(q.keywords[1], "AGN")
        self.assertTrue(q.will_or_keywords())

        qurl = q.getqueryurl()
        self.assertTrue("+OR+%28title+" in qurl)

        import pytest
        self.assertRaises(ValueError, 
                          self.srv.create_query, orkw=["galaxy", "AGN"])
        

class RegQueryTest(unittest.TestCase):
    
    def testCtor(self):
        self.q = reg.RegistryQuery()

    def testWaveband(self):
        self.testCtor()
        self.assertTrue(self.q.waveband is None)
        self.q.waveband = "Radio"
        self.assertTrue(self.q.waveband is not None)
        self.assertEqual(self.q.waveband, "Radio")

        del self.q.waveband 
        self.assertTrue(self.q.waveband is None)

        self.q.waveband = "radio"
        self.assertTrue(self.q.waveband is not None)
        self.assertEqual(self.q.waveband, "Radio")

        self.q.waveband = "millimeter"
        self.assertEqual(self.q.waveband, "Millimeter")
        self.q.waveband = "infrared"
        self.assertEqual(self.q.waveband, "Infrared")
        self.q.waveband = "ir"
        self.assertEqual(self.q.waveband, "Infrared")
        self.q.waveband = "optical"
        self.assertEqual(self.q.waveband, "Optical")
        self.q.waveband = "uv"
        self.assertEqual(self.q.waveband, "UV")
        self.q.waveband = "UV"
        self.assertEqual(self.q.waveband, "UV")
        self.q.waveband = "euv"
        self.assertEqual(self.q.waveband, "EUV")
        self.q.waveband = "x-ray"
        self.assertEqual(self.q.waveband, "X-ray")
        self.q.waveband = "gamma-ray"
        self.assertEqual(self.q.waveband, "Gamma-ray")

        try:
            self.q.waveband = "blob"
            self.fail("Failed to reject unrecognized waveband")
        except ValueError:
            pass

    def testServicetype(self):
        self.testCtor()
        self.assertTrue(self.q.servicetype is None)
        self.q.servicetype = "catalog"
        self.assertTrue(self.q.servicetype is not None)
        self.assertEqual(self.q.servicetype, "catalog")

        del self.q.servicetype
        self.assertTrue(self.q.servicetype is None)

        self.q.servicetype = "scs"
        self.assertEqual(self.q.servicetype, "scs")
        self.q.servicetype = "ssa"
        self.q.servicetype = "conesearch"
        self.q.servicetype = "Conesearch"
        self.q.servicetype = "table"
        self.q.servicetype = "image"
        self.q.servicetype = "sia"
        self.q.servicetype = "spectrum"
        self.q.servicetype = "spectra"
        self.q.servicetype = "ssa"
        self.q.servicetype = "ssap"
        self.q.servicetype = "line"
        self.q.servicetype = "sla"
        self.q.servicetype = "slap"
        self.q.servicetype = "simpleImageAccess"
        self.q.servicetype = "SimpleImageAccess"
        self.q.servicetype = "simpleSpectralAccess"
        self.q.servicetype = "SimpleSpectralAccess"

        try:
            self.q.servicetype = "SimpleGooberAccess"
            self.fail("failed to reject unrecognized servicetype")
        except ValueError:
            pass
        
    def testOrKeywords(self):
        self.testCtor()
        self.assertTrue(not self.q.will_or_keywords())
        self.q.or_keywords(False)
        self.assertTrue(not self.q.will_or_keywords())
        self.q.or_keywords(True)
        self.assertTrue(self.q.will_or_keywords())

    def testKeywords(self):
        self.testCtor()
        self.assertEqual(self.q.keywords, [])

        self.q.addkeywords("galaxy")
        self.assertEqual(self.q.keywords, ["galaxy"])

        self.q.clearkeywords()
        self.assertEqual(self.q.keywords, [])
        
        self.q.addkeywords("AGN")
        self.assertEqual(self.q.keywords, ["AGN"])
        self.q.addkeywords("galaxy")
        self.assertEqual(self.q.keywords, ["AGN", "galaxy"])
        self.q.addkeywords("dwarf companion")
        self.assertEqual(self.q.keywords, ["AGN", "galaxy", "dwarf companion"])
        self.q.addkeywords("HI velocities".split())
        self.assertEqual(self.q.keywords, 
                       ["AGN", "galaxy", "dwarf companion", "HI", "velocities"])

        self.q.removekeywords("HI galaxy".split())
        self.assertEqual(self.q.keywords, 
                          ["AGN", "dwarf companion", "velocities"])

    def testPredicates(self):
        self.testCtor()
        self.assertEqual(self.q.predicates, [])
        
        self.q.addpredicate("Identifier like 'ivo://ncsa.%'")
        self.assertEqual(self.q.predicates, ["Identifier like 'ivo://ncsa.%'"])
        self.q.addpredicate("publisher like '%nrao%'")
        self.assertEqual(self.q.predicates, 
                          ["Identifier like 'ivo://ncsa.%'",
                           "publisher like '%nrao%'"])

        self.q.removepredicate("Identifier like 'ivo://ncsa.%'")
        self.assertEqual(self.q.predicates, ["publisher like '%nrao%'"])

        self.q.clearpredicates()
        self.assertEqual(self.q.predicates, [])

        try:
            self.q.predicates = "Identifier like 'ivo://ncsa.%'"
            self.fail("predicates property is not read-only")
        except AttributeError:
            pass

        self.q.addpredicate("Identifier like 'ivo://ncsa.%'")
        self.q.predicates[0] = "publisherid like 'ivo://ncsa.%'"
        self.assertEqual(self.q.predicates, ["Identifier like 'ivo://ncsa.%'"])

    def testKeywordsToPred(self):
        kws = "gal AGN"
        self.testCtor()

        pat = "(title LIKE '%{t}%' OR shortName LIKE '%{t}%' OR " + \
              "identifier LIKE '%{t}%' OR " + \
              "[content/subject] LIKE '%{t}%' OR "+\
              "[curation/publisher] LIKE '%{t}%' OR " + \
              "[content/description] LIKE '%{t}%')"

        pred = self.q.keywords_to_predicate([kws], True)
        self.assertEqual(pred, pat.format(**{"t": kws}))

        kws = kws.split()
        pred = self.q.keywords_to_predicate(kws[0:1], False)
        self.assertEqual(pred, pat.format(**{"t": kws[0]}))

        pred = self.q.keywords_to_predicate(kws, True)
        self.assertEqual(pred, 
                          (pat.format(**{"t": kws[0]})) + " OR " + 
                          (pat.format(**{"t": kws[1]})))

        pred = self.q.keywords_to_predicate(kws, False)
        self.assertEqual(pred, 
                          (pat.format(**{"t": kws[0]})) + " AND " + 
                          (pat.format(**{"t": kws[1]})))


    def testCreateURL(self):
        baseurl = reg.RegistryService.STSCI_REGISTRY_BASEURL + \
            "VOTCapBandPredOpt?VOTStyleOption=2&"

        self.testCtor()
        self.q.waveband = "ir"
        qurl = self.q.getqueryurl()
        self.assertTrue(qurl.startswith(baseurl))
        self.assertTrue("&waveband=Infrared" in qurl)

        self.q.servicetype = "image"
        qurl = self.q.getqueryurl()
        self.assertTrue("&waveband=Infrared" in qurl)
        self.assertTrue("&capability=SimpleImageAccess" in qurl)
        self.assertTrue(qurl.endswith("&predicate="))

        self.q.addpredicate("publisher like '%nrao%'")
        qurl = self.q.getqueryurl()
        self.assertTrue("&predicate=%28publisher+like+%27%25nrao%25%27%29" in qurl,
                     "unexpected predicate: " + qurl)
        self.q.addpredicate("Identifier like 'ivo%'")
        qurl = self.q.getqueryurl()
        self.assertTrue("&predicate=%28publisher+like+%27%25nrao%25%27%29+AND+%28Identifier+like+%27ivo%25%27%29" in qurl,
                     "unexpected predicate: " + qurl)
        self.q.clearpredicates()
        qurl = self.q.getqueryurl()
        self.assertTrue(qurl.endswith("&predicate="))

        # test keywords; not that we tested the pre-URL-encoding of the keyword
        # constraints in testKeywordsToPred()
        self.q.addkeywords(["galaxy", "AGN"])
        qurl = self.q.getqueryurl()
        self.assertTrue("&waveband=Infrared" in qurl)
        self.assertTrue("&capability=SimpleImageAccess" in qurl)
        self.assertTrue("&predicate=1" not in qurl)
        for kw in "AGN galaxy".split():
            for col in ["title", "shortName", "identifier", 
                        "%5Bcontent%2Fsubject%5D", "%5Bcuration%2Fpublisher%5D", 
                        "%5Bcontent%2Fdescription%5D"]:
                constraint = col + "+LIKE+%27%25" + kw + "%25%27"
                self.assertTrue(constraint in qurl, 
                         "Missing {0} constraint:\n{1}".format(constraint,qurl))
        self.assertTrue("+OR+shortName+" in qurl)
        self.assertTrue("+AND+%28title+" in qurl,
                     "unexpected predicate: " + qurl)
        self.q.or_keywords(True)
        qurl = self.q.getqueryurl()
        self.assertTrue("+OR+%28title+" in qurl)

        self.q.addpredicate("publisher like '%nrao%'")
        qurl = self.q.getqueryurl()
        self.assertTrue("&predicate=%28publisher+like+%27%25nrao%25%27%29+AND+%28" 
                     in qurl,
                     "unexpected predicate: " + qurl)

class RegResultsTest(unittest.TestCase):

    def setUp(self):
        # load a canned result file
        resultfile = get_pkg_data_filename(regresultfile)
        self.tbl = votableparse(resultfile)

    def testCtor(self):
        self.r = reg.RegistryResults(self.tbl)
        self.assertTrue(self.r.votable is not None)
        self.assertEqual(self.r.nrecs, 4)

    def testGetValue(self):
        self.testCtor()
        v = self.r.getvalue("shortName", 0)
        if sys.version_info[0] >= 3:
            self.assertTrue(isinstance(v, bytes))
        else:
            self.assertTrue(isinstance(v, str))
        self.assertEqual(v, b"J/MNRAS/333/100 [1]")

        v = self.r.getvalue("subject", 0)
        self.assertTrue(isinstance(v, tuple))
        self.assertEqual(len(v), 4)
        if sys.version_info[0] >= 3:
            self.assertTrue(isinstance(v[0], bytes))
        else:
            self.assertTrue(isinstance(v[0], str))
        self.assertEqual(v[0], b"AGN")
        self.assertEqual(v[1], b"Galaxies")
        self.assertEqual(v[3], b"Spectroscopy")

    def testListVal(self):
        self.testCtor()
        self.assertTrue(isinstance(self.r.getvalue("subject", 0), tuple))
        self.assertTrue(isinstance(self.r.getvalue("waveband", 0), tuple))
        self.assertTrue(isinstance(self.r.getvalue("contentLevel", 0), tuple))
        self.assertTrue(isinstance(self.r.getvalue("type", 0), tuple))

    def testGetRecord(self):
        self.testCtor()
        for i in range(4):
            rec = self.r.getrecord(0)
            self.assertTrue(isinstance(rec, reg.SimpleResource))
        

class SimpleResTest(unittest.TestCase):

    def setUp(self):
        resultfile = get_pkg_data_filename(regresultfile)
        self.tbl = votableparse(resultfile)
        self.result =  reg.RegistryResults(self.tbl)
        self.setrec(0)

    def setrec(self, i):
        self.r = self.result.getrecord(i)

    def testArbitraryCol(self):
        self.setrec(0)
        self.assertTrue(self.r["description"].startswith(b"We use redshift determinations and spectral"))

    def testAttr(self):
        self.assertEqual(self.r.shortname, b"J/MNRAS/333/100 [1]")
        self.assertEqual(self.r.title, 
                          b"Radio galaxies in the 2dFGRS (Magliocchetti+, 2002)")
        self.assertEqual(self.r.ivoid, b"ivo://CDS.VizieR/J/MNRAS/333/100#1")
        self.assertEqual(self.r.accessurl, 
           "http://vizier.u-strasbg.fr/cgi-bin/VizieR-2?-source=J/MNRAS/333/100")
        self.assertEqual(self.r.publisher, b"CDS")
        self.assertEqual(self.r.tags, b"Web Page")

        # pdb.set_trace()
        self.assertTrue(isinstance(self.r.subject, tuple))
        self.assertEqual(len(self.r.subject), 4)
        self.assertEqual(self.r.subject[0], b"AGN")
        self.assertEqual(self.r.subject[1], b"Galaxies")
        self.assertEqual(self.r.subject[3], b"Spectroscopy")

        self.assertEqual(self.r.type[0], b"Catalog")
        self.assertEqual(self.r["interfaceClass"], b"WebBrowser")
        self.assertEqual(self.r.standardid, b"")
        self.assertEqual(self.r.capability, b"")

    def testListVal(self):
        self.assertTrue(isinstance(self.r.subject, tuple))
        self.assertTrue(isinstance(self.r.waveband, tuple))
        self.assertTrue(isinstance(self.r.type, tuple))
        self.assertTrue(isinstance(self.r.contentlevel, tuple))

    def testCap(self):
        self.setrec(2)
        self.assertEqual(self.r.tags, b"Catalog")
        self.assertEqual(self.r.type[0], b"Catalog")
        self.assertEqual(self.r.standardid, b"ivo://ivoa.net/std/ConeSearch")
        self.assertEqual(self.r.capability, b"ConeSearch")

        self.setrec(3)
        self.assertEqual(self.r.tags, b"Images")
        self.assertEqual(self.r.type[0], b"Archive")
        self.assertEqual(self.r.standardid, b"ivo://ivoa.net/std/SIA")
        self.assertEqual(self.r.capability, b"SimpleImageAccess")


__all__ = "RegServiceTest RegQueryTest RegResultsTest SimpleResTest".split()
def suite():
    tests = []
    for t in __all__:
        tests.append(unittest.makeSuite(globals()[t]))
    return unittest.TestSuite(tests)

if __name__ == "__main__":
    unittest.main()
